// Generated by the protocol buffer compiler.  DO NOT EDIT!
// source: EnumRoot.proto

package io.confluent.kafka.serializers.protobuf.test;

public final class EnumRootOuter {
  private EnumRootOuter() {}
  public static void registerAllExtensions(
      com.google.protobuf.ExtensionRegistryLite registry) {
  }

  public static void registerAllExtensions(
      com.google.protobuf.ExtensionRegistry registry) {
    registerAllExtensions(
        (com.google.protobuf.ExtensionRegistryLite) registry);
  }
  /**
   * Protobuf enum {@code EnumRoot}
   */
  public enum EnumRoot
      implements com.google.protobuf.ProtocolMessageEnum {
    /**
     * <code>HELLO = 0;</code>
     */
    HELLO(0),
    /**
     * <code>GOODBYE = 1;</code>
     */
    GOODBYE(1),
    UNRECOGNIZED(-1),
    ;

    /**
     * <code>HELLO = 0;</code>
     */
    public static final int HELLO_VALUE = 0;
    /**
     * <code>GOODBYE = 1;</code>
     */
    public static final int GOODBYE_VALUE = 1;


    public final int getNumber() {
      if (this == UNRECOGNIZED) {
        throw new java.lang.IllegalArgumentException(
            "Can't get the number of an unknown enum value.");
      }
      return value;
    }

    /**
     * @param value The numeric wire value of the corresponding enum entry.
     * @return The enum associated with the given numeric wire value.
     * @deprecated Use {@link #forNumber(int)} instead.
     */
    @java.lang.Deprecated
    public static EnumRoot valueOf(int value) {
      return forNumber(value);
    }

    /**
     * @param value The numeric wire value of the corresponding enum entry.
     * @return The enum associated with the given numeric wire value.
     */
    public static EnumRoot forNumber(int value) {
      switch (value) {
        case 0: return HELLO;
        case 1: return GOODBYE;
        default: return null;
      }
    }

    public static com.google.protobuf.Internal.EnumLiteMap<EnumRoot>
        internalGetValueMap() {
      return internalValueMap;
    }
    private static final com.google.protobuf.Internal.EnumLiteMap<
        EnumRoot> internalValueMap =
          new com.google.protobuf.Internal.EnumLiteMap<EnumRoot>() {
            public EnumRoot findValueByNumber(int number) {
              return EnumRoot.forNumber(number);
            }
          };

    public final com.google.protobuf.Descriptors.EnumValueDescriptor
        getValueDescriptor() {
      return getDescriptor().getValues().get(ordinal());
    }
    public final com.google.protobuf.Descriptors.EnumDescriptor
        getDescriptorForType() {
      return getDescriptor();
    }
    public static final com.google.protobuf.Descriptors.EnumDescriptor
        getDescriptor() {
      return io.confluent.kafka.serializers.protobuf.test.EnumRootOuter.getDescriptor().getEnumTypes().get(0);
    }

    private static final EnumRoot[] VALUES = values();

    public static EnumRoot valueOf(
        com.google.protobuf.Descriptors.EnumValueDescriptor desc) {
      if (desc.getType() != getDescriptor()) {
        throw new java.lang.IllegalArgumentException(
          "EnumValueDescriptor is not for this type.");
      }
      if (desc.getIndex() == -1) {
        return UNRECOGNIZED;
      }
      return VALUES[desc.getIndex()];
    }

    private final int value;

    private EnumRoot(int value) {
      this.value = value;
    }

    // @@protoc_insertion_point(enum_scope:EnumRoot)
  }


  public static com.google.protobuf.Descriptors.FileDescriptor
      getDescriptor() {
    return descriptor;
  }
  private static  com.google.protobuf.Descriptors.FileDescriptor
      descriptor;
  static {
    java.lang.String[] descriptorData = {
      "\n\016EnumRoot.proto*\"\n\010EnumRoot\022\t\n\005HELLO\020\000\022" +
      "\013\n\007GOODBYE\020\001B=\n,io.confluent.kafka.seria" +
      "lizers.protobuf.testB\rEnumRootOuterb\006pro" +
      "to3"
    };
    descriptor = com.google.protobuf.Descriptors.FileDescriptor
      .internalBuildGeneratedFileFrom(descriptorData,
        new com.google.protobuf.Descriptors.FileDescriptor[] {
        });
  }

  // @@protoc_insertion_point(outer_class_scope)
}
