/*
 * Copyright 2020 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.kafka.schemaregistry.exceptions;

public class SchemaVersionNotSoftDeletedException extends SchemaRegistryException {
  private final String subject;
  private final String version;

  public SchemaVersionNotSoftDeletedException(String subject, String version) {
    super("Subject: " + subject + " Version: "
            + version + " must be deleted first before being permanently deleted.");
    this.subject = subject;
    this.version = version;
  }

  public String getSubject() {
    return subject;
  }

  public String getVersion() {
    return version;
  }
}
